/*********************************************************************/
/* neblo.h                                                           */
/* types, constants, and prototypes for neblo                        */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#ifndef NEBLO_API_H
#define NEBLO_API_H


/* max size for a neblo message payload */

#define NEBLO_MSGSIZE   4096  /* HEADERSIZE_F + NEBLO_MSGSIZE <= KEYBUFLEN */


typedef u_int64_t neblo_addr;
typedef u_int16_t neblo_port;


/*  Handlers for messages  */

typedef int ( *handler ) ( neblo_port, neblo_addr, void *, unsigned int * );


/*  Handlers for peer arrival/departures  */

typedef int ( *manager ) ( neblo_addr, neblo_addr );


/*
 *  Handles (not handler!) for request-response msgs,
 *  to wait/test for responses
 */

typedef int neblo_rhandle;
#define R_COMPLETE ((neblo_rhandle)(-1))
#define R_FAILED   ((neblo_rhandle)(-2))


/*  Initialization routine  */

void neblo_init ( int argc, char **argv );


/*
 *  Attach a handler to a port, to handle UniDirectional msgs at a recipient
 *  or at a forwarder node.
 *  The originator is also a forwarder, if it is not also the recipient.
 */

void neblo_set_receiver_handler_UD ( neblo_port, handler );

void neblo_set_forwarder_handler_UD ( neblo_port, handler );


/*
 *  Attach a handler to a port, to handle REQuest-RESponse msgs at a recipient
 *  or at a forwarder.
 *  The originator is also a forwarder, if it is not also the recipient.
 *  At the recipient, the handler should compute an answer so "msg_len" might
 *  change (actually it is only allowed to shrink).
 */

void neblo_set_receiver_handler_REQ ( neblo_port, handler );

void neblo_set_forwarder_handler_REQ ( neblo_port, handler );

void neblo_set_forwarder_handler_RES ( neblo_port, handler );


/*
 *  Set the migration and regeneration handlers,
 *  to manage node arrivals/departures.
 */

void neblo_set_refresh_handler ( manager );

void neblo_set_clean_handler ( manager );


/*  Connect to the overlay, after having set up the handlers.  */

int neblo_connect ( void );


/*  Send a UniDirectional msg.  No answer will be provided by the recipient  */

int neblo_UD_send
(
  unsigned int num_copies, neblo_port port, neblo_addr addr, void *msg,
  unsigned int msg_len
);


/*  A variant of neblo_UD_send(), for exclusive use in handlers  */

int neblo_UD_send_h
( neblo_port port, neblo_addr addr, void *msg, unsigned int msg_len );


/*
 *  Send a REQuest-RESponse msg.  An answer will be provided by the recipient.
 *  Parm "maxlen" denotes both this msg len and the max len allowed for the
 *  expected answer.
 *  Semantics is non-blocking with a split-phase architecture, so wait/test
 *  routines must be invoked to synchronize with the answer; parm "rhandle"
 *  binds the two phases.
 */

int neblo_REQ_send
(
  unsigned int num_copies, neblo_port port, neblo_addr addr, void *msg,
  unsigned int maxlen, neblo_rhandle *rhandle
);

/*  A variant of neblo_REQ_send(), for exclusive use in handlers  */

int neblo_REQ_send_h
(
  neblo_port port, neblo_addr addr, void *msg,
  unsigned int maxlen, neblo_rhandle *rhandle
);


int neblo_RES_wait ( neblo_rhandle );

int neblo_RES_test ( neblo_rhandle );


/*  Miscellaneous  */

unsigned int neblo_in_range ( neblo_addr,  neblo_addr,  neblo_addr );

void neblo_lock( void );

int neblo_trylock( void );

void neblo_unlock( void );
