#ifndef NEBLO_H
#define NEBLO_H


/*
 *  routing terms:
 *
 *  "distance":  each peer owns an interval of keys.  Intervals of distinct
 *               peers never overlap.  Thus, an interval of keys can be denoted
 *               by its lower bound.  The distance between peers is the
 *               difference between their key interval lower bounds.  The
 *               distance of a key from a peer is the difference between the
 *               key and the lower bound of that peer's interval.
 *  "neighbour": a peer located immediately right of local peer along the
 *               ring ("successor"), or a neighbour of the successor
 *               within MAXNEIGH hops from local peer.
 *               More than one neighbours are allowed, to make the ring
 *               topology robust against peer failures.
 *               Neighbours are always consecutive, in the ring topology.
 *               Each peer has one outgoing conn towards each of its
 *               neighbours.  The successor also has a backward connection to
 *               local peer; that is, local peer and its successor share a
 *               bidirectional connection, useful to "travel backward" in the
 *               ring.
 *  "finger":    a peer located at usually a greater distance than neighbours,
 *               not exceeding MAXFNGDIST.
 *               More that one fingers are allowed.
 *               Fingers make routing faster, but also more robust against
 *               failures of successor.
 *               Fingers are located at increasing distances, according to
 *               a progression recursively defined in finger_distance_rule()
 *               (see file neblo.c).  The progression is usually geometrical,
 *               to allow logN routing.  The distance to the nearest finger
 *               must be <= CUTOFF; it is convenient that such distance be
 *               as close to the CUTOFF limit as possible.
 *               Each peer has one outgoing conn towards each of its fingers.
 *  "friend":    a neighbour, a finger, or any peer who has the local peer as
 *               neighbour or finger.
 *               More than one friends are allowed, as a consequence.
 *               Friends are kept for security purposes: they serve to check
 *               that protocol commands only arrive through established
 *               routing connections, so as to exclude spoofing.
 *               Each peer sees each of its friends as one incoming connection.
 *               One of the friends corresponds to the "predecessor" in the
 *               ring.  Another one corresponds to the backward connection from
 *               the successor.
 *
 *  CUTOFF, MAXNEIGH, MAXFNGDIST, neighbours, fingers, etc. are defined below.
 *  Typical value for CUTOFF is 2^16.
 *  Typical value for MAXFNGDIST is half the ring lenght (0x8000000000000 in
 *  the current prototype).
 *
 */

#define VERBOSE

#undef DEBUG
#undef DEBUG_J                /* Join a ring                           */
#undef DEBUG_F                /* establish a Friendship                */
#undef DEBUG_M                /* Maintain routing tables               */
#undef DEBUG_R                /* Routing table printouts               */
#undef DEBUG_RW               /* Read/Write/Data/Refresh messages      */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <signal.h>             /* to catch/mask signals                */
#include <errno.h>              /* errno variable                       */
#include <openssl/rsa.h>	/* various crypto functions             */
#include <openssl/ripemd.h>	/* RIPEMD160 crypto hash function       */
#include <openssl/blowfish.h>   /* symmetric stream cipher              */
#include <openssl/rand.h>       /* cryptographic random generator       */
#include <sys/types.h>		/* standard system types		*/
#include <netinet/in.h>		/* Internet address structures          */
#include <sys/socket.h>		/* socket interface functions           */
#include <netdb.h>		/* host to IP resolution                */
#include <sys/time.h>		/* timeout values, gettimeofday() etc.  */
#include <unistd.h>		/* for table size calculations          */
#include <string.h>		/* "strcmp", etc.                       */
#include <pthread.h>

#include <neblo.h>


#define COMBUF_SIZE  65536      /* size of low-level ciphred comm. bufs */
#define BUFLEN       100        /* size of many strings and buffers;    */
				/* if this is to change, check for      */
                                /* possible %100s in fscanf() format    */
                                /* strings.                             */
#define SLACK        28         /* BUFLEN+SLACK = 2**k && SLACK != 0    */
#define LINESIZE     60000      /* max size of lines in startup files;  */
                                /* if this is to change, see below.     */
#define KEYBUFLEN    60000      /* KEYBUFSIZE >= LINESIZE; if this is   */
                                /* change, check for possible %50000s   */
                                /* in fscanf() format strings.          */

#define CUTOFF_LOG   16
#define CUTOFF       ((neblo_addr)(1<<CUTOFF_LOG))

#define MAXFNGDIST   (~(((neblo_addr)(-1)) >> 1))
                                /* dist. of fingers is <= MAXFNGDIST    */
                                /* typically half the ring size:        */
                                /* 0x8000000000000000                   */

#define MAXFNGMAGN   (sizeof(neblo_addr) - CUTOFF_LOG)

#define MAXNEIGH     10         /* max no. of neighbours                */

                                /* NUMFINGERS, the number of allowed    */
                                /* fingers, is a variable initialized   */
                                /* by neblo_init()                      */

#define MAXFRIENDS   (MAXNEIGH+NUMFINGERS)
                                /* max no. of friends                   */

#define MAX_CP       5          /* max no. of copies of msgs            */

#define BACKLOG      0          /* for the listen()                     */
#define TIMEOUT      10         /* timeout (secs) on most sock read()s  */
#define SERV_TIMEOUT (10*CHK_PERIOD_NEIGHS)
                                /* 10 * max time between keepalives     */
                                /* for inbound connection (friends).    */
#define NOTIMEOUT    -1         /* no timeout on some socket read()s    */

#define REC_CMD_DEADLINE 30     /* 30 sec exp. time for recursive cmds  */

#define HEADERSIZE_W 54         /* size of headers of UD messages:      */
/* "write 0123 01234567 0123456789ABCDEF 0123456789ABCDEF "             */
/*        port len      addr (hex)       real_addr (hex)                */

#define HEADERSIZE_R 62         /* size of headers of REQ and RES msgs: */
/* "read 0123 01234567 01234567 0123456789ABCDEF 0123456789ABCDEF "     */
/*       port len      handle   addr (hex)       real_addr (hex)        */
/* "data 0123 01234567 01234567 0123456789ABCDEF 0123456789ABCDEF "     */
/*       port len      handle   addr (hex)       real_addr (hex)        */

#define HEADERSIZE_F 90         /* size of headers of "refresh" msgs:   */
/* "refr 0123456789ABCDEF 0123456789ABCDEF 0123456789ABCDEF             */
/*       0123456789ABCDEF 0123456789ABCDEF "                            */
/*       lo_key (hex)     up_key (hex)     lo_addr (hex)                */
/*       up_addr (hex)    req_key(hex)                                  */

#define DEFAULT_CFG_FILE     "neblorc"
#define DEFAULT_SEEDS_FILE   "nebloseeds"



/* Data types */

typedef enum { WRITE,READ,DATA,REFR } rwd_t;

typedef struct                  /* a ciphred socket */
{
  int sock;                     /* the socket */
  unsigned char *ses_k;         /* room for the seed of Blowfish cipher */
  BF_KEY *bf_ses_k;             /* room for the Blowfish key */
  unsigned char ivec[8];        /* Blowfish needs this, but we don't use it */
  int num;                      /* Blowfish needs this */
  char *msg;                    /* buffer for the "howdy" initial set up */
  unsigned int combuf_len;      /* len of msg stored into combuf (below) */
  char *combuf;                 /* buffer for actual communications */
} csocket_t;

typedef struct                  /* a seed:  first contact when joining ring */
{
  struct in_addr addr;
  unsigned int port;
  RSA *pubkey;
} seed_t;

typedef struct                  /* handles for cmds whose answers arrive  */
                                /* after running through a chain of peers */
                                /* (recursive-style cmds)                 */
{
  pthread_mutex_t lock;
  u_int32_t       val;
  struct timeval  timestamp;
} cmd_handle_t;

typedef enum                    /* possible statuses of a remote peer     */
{
    NONE=0, DEAD=0x01, TOCALL=0x02, PROBABLYDEAD=0x04,
    OK=0x80
} status_t;

typedef struct pt               /*  descriptor of a remote peer  */
{
  pthread_mutex_t lock;         /* mutex between init_friendship() and      */
                                /* garb coll done by check_neighbours().    */
                                /* also to protect "dialogs", see comments  */
                                /* below.                                   */
  struct in_addr addr;
  u_int16_t port;
  u_int16_t dummy;
  RSA *pubkey;

  status_t status;              /* OK, DEAD, TOCALL, etc */

  csocket_t *cs;                /* routing/mainten. conn towards this peer,  */
                                /* when in neighbour list/finger tables      */

  csocket_t **in_cs;            /* ptr to routing/maintenance conn incoming  */
                                /* from this peer (friend list only)         */
/*
 *  NOTE:  commands from the connect protocol will travel along temp conns,
 *  not the above ones (which can only be established after having joined the
 *  ring).
 *  NOTE:  some commands from the maintenance protocol may require multiple
 *  exchanges of messages among partners along routing connections.
 *  These commands are:  "neighlist" and "findsucc" ("newsucc" and "welcome"
 *  occurs on tmp connections).  To protect the "dialogs" issued by these
 *  commands against other cmds issued by other threads on the same routing
 *  connections, we must make explicit use of the lock struct inside this
 *  struct.
 */

  cmd_handle_t handle;          /* in finger table:                          */
                                /* handle for a "findfinger" recursive req   */
                                /* issued by local peer.                     */
                                /* in neigh list, as a dead successor:       */
                                /* handle for an in progress "findsucc" req  */
                                /* issued by local peer.                     */
  unsigned int magn;            /* in friend list: magnitude of a            */
                                /* "findfinger" req issued to local peer by  */
                                /* this remote peer.                         */
  unsigned int deadline;        /* decr at each failed communic, until DEAD  */
  u_int8_t from_predecessor;    /* in friend list: 1 iff conn from predcssor */
  u_int8_t from_successor;      /* in friend list: 1 iff conn from successor */
  u_int8_t self;
  u_int8_t tmpunavail;
  struct pt *next;
} peer_t;

typedef struct                  /* the local peer as member of the ring      */
{
  neblo_addr my_id_src;         /* source of the "unique" id (below)         */
  neblo_addr my_id;             /* this peer's "unique" id;                  */
                                /* computed as RIPEMD160 of source (above),  */
                                /* XOR-folded over u_int64_t                 */
  neblo_addr up_bound;          /* intvl of locally stored keys: upper bound */
                                /* (lower bound is my_id itself).            */
  neblo_addr smoke;             /* see findfinger.c to understand this       */
  unsigned int enabled;         /* 1 iff this peer has joined the ring       */
  peer_t *neigh_list;           /* routing table: neighbour list             */
  peer_t *neigh_list_tail;
  peer_t *finger_table;         /* routing table: finger table (array)       */
  peer_t *friend_list;          /* routing table: friend list                */
  peer_t *friend_list_tail;
  pthread_mutex_t lock;         /* lock on the entire struct; ensures        */
                                /* consistent access to "up_bound" and       */
                                /* routing tables.                           */
  pthread_mutex_t msg_lock;     /* mutual excl between neblo_join() and msgs */
  pthread_mutex_t join_lock;    /* prevents neblo_join() from reentering     */
                                /* only used when (re-)joining the ring      */
  cmd_handle_t join_handle;     /* for the recursive-style "register" cmd.   */
                                /* only used when (re-)joining the ring      */

} ring_t;

typedef struct ptptr
{
  peer_t *f;
  struct ptptr *next;
} peerptr_t;

typedef struct
{
  ring_t ring;
  peerptr_t *peerlist;
} ringofptrs_t;

typedef struct                  /* an incoming connection carrying requests */
{
  char *buf;			/* room to store arriving requests */
  csocket_t *csock;             /* requests come from a remote peer */
  struct in_addr peer_addr;     /* addr of the remote peer issuing the reqst */
  pthread_mutex_t mutx;		/* mutex associated to sem */
  pthread_cond_t sem;		/* to block/unblock the associated thread */
  pthread_t th_id;		/* id. of the associated serving thread */
  void *next;
} reqst_t;



/* Extern globals */

extern u_int16_t        PORT;           /* default peer port                */
extern unsigned int     MAXCONNS;       /* default max no. of active conns  */
extern unsigned int     MAXSEEDS;       /* no. of seeds locally available   */
extern struct in_addr   my_ipaddr;      /* IP addr of this host             */
extern struct in_addr   my_ipaddr_alt;  /* IP addr of "localhost",          */
                                        /* different from the one above.    */
extern seed_t          *seed;           /* seeds from local seed file       */
extern ring_t          *ring;           /* peer ring                        */

extern unsigned int     NUMFINGERS;     /* actual val depend on the finger  */
                                        /* distance law, see neblo.c        */

/* Period (secs) of rout tabl maint:  neighbours, and fingers */
extern unsigned int     CHK_PERIOD_NEIGHS;
extern unsigned int     CHK_PERIOD_FINGERS;

/* After this no. of attempts, a neighbour is considered DEAD               */
extern unsigned int     NEIGH_DEADLINE;

/* Chk period and deadline (secs) for expected requests */
extern unsigned int     EXPD_PERIOD;
extern unsigned int     EXPD_DEADLINE;

/* Chk period and deadline (secs) for pending read messages */
extern unsigned int     PEND_READ_PERIOD;
extern unsigned int     PEND_READ_DEADLINE;

/* Chk period, and deadline (secs) for delayed msgs */
extern unsigned int     RETRY_PERIOD;
extern unsigned int     RETRY_DEADLINE;

/*
 * Queue of descriptors of incoming connections.
 * A connection persists until one of the partners decides to shut it down,
 * for whatever reason.
 * Once established, an incoming conn carries remote requests to be served
 * by local peer.
 */
extern reqst_t         *reqst;          /* descriptor queue */
extern reqst_t         *reqst_head;
extern reqst_t         *reqst_tail;
extern pthread_mutex_t  reqst_lock;     /* mutex on reqsts list */
extern pthread_cond_t   reqst_sem;      /* cond on empty reqsts list */

extern int              wk_s;           /* to wake up sleeping main thread   */
extern RSA             *pubkey;         /* (ptr to) this peer's public key   */
extern char            *pubkeymod;      /* holds modulus of pubkey           */
extern char            *pubkeyexp;      /* holds exponent of pubkey          */
extern unsigned int     pubkey_size;    /* RSA_size(public key)              */
extern RSA             *seckey;         /* (ptr to) this peer's secret key   */
extern FILE            *skf;            /* secret key file, open by init()   */
extern char            *seckeybuf;      /* buf for reading seckey from file  */
extern pthread_mutex_t  malloc_lock;    /* mutex on malloc() as well as      */
                                        /* every other lib routine using it  */
/*
 * mutex between services that make use of the routing tables, on one hand,
 * and garbage collector of the routing tables, on the other hand.
 */
extern pthread_mutex_t  clean_lock;
extern unsigned int     dont_clean;

/* arrays of application-programmed handlers, and related stuff */

extern pthread_mutex_t  __receiver_handler_lock;
extern handler         *__receiver_handler_UD;
extern handler         *__receiver_handler_REQ;
extern neblo_addr       __refresh_addr;
extern unsigned int     __refresh_in_progress;
extern manager          __refresh_handler;
extern manager          __clean_handler;
extern pthread_mutex_t  __forwarder_handler_lock;
extern handler         *__forwarder_handler_UD;
extern handler         *__forwarder_handler_REQ;
extern handler         *__forwarder_handler_RES;



/* Function prototypes */

int   neblo_submit ( void );
void  xor_fold ( int8_t *, unsigned int, int8_t *, unsigned int );

void *do_serv ( void * );

/*  These are the protocol cmds  */
int     neblo_announce ( reqst_t * );
int     neblo_findfinger ( reqst_t * );
int     neblo_findsucc ( reqst_t * );
int     neblo_friendship ( reqst_t * );
int     neblo_handle_UD_msgs ( reqst_t *, unsigned int );
int     neblo_handle_REQ_msgs ( reqst_t *, unsigned int );
int     neblo_handle_RES_msgs ( reqst_t * );
int     neblo_handle_refresh_msgs ( reqst_t *, unsigned int );
int     neblo_join ( reqst_t * );
int     neblo_neighlist ( reqst_t * );
int     neblo_newfinger ( reqst_t * );
int     neblo_newsucc ( reqst_t * );
int     neblo_register ( reqst_t * );
int     neblo_welcome ( reqst_t * );

void    finger_create ( pthread_t );
void   *neblo_check_neighbours ( void * );
void   *neblo_check_fingers ( void * );
void    ask_refresh_addrs ( neblo_addr, neblo_addr, unsigned int, pthread_t );
int     republish_addrs ( neblo_addr, neblo_addr, neblo_addr, unsigned int );
int     clean_addrs ( neblo_addr, neblo_addr, unsigned int );

void    clean_peer_desc ( peer_t * );
void    clean_peer_desc_nolock ( peer_t * );
int     call_peer ( peer_t *, unsigned int, int, pthread_t );
int     tell_peer ( peer_t *, void *, unsigned int, int, pthread_t );
int     tell_peer_nolock ( peer_t *, void *, unsigned int, int, pthread_t );
int     listen_peer ( peer_t *,void *, unsigned int, long, int, pthread_t );
int     listen_peer_nolock
          ( peer_t *, void *, unsigned int, long, int, pthread_t );
void    append_peer_list ( peer_t **, peer_t **, peer_t * );
void    clean_peer_list ( peer_t **, peer_t ** );
void    purge_peer_list ( peer_t **, peer_t ** );
peer_t *is_friend ( csocket_t * );
peer_t *predecessor ( void );
peer_t *successor ( void );
peer_t *from_predecessor ( csocket_t * );
peer_t *from_successor ( csocket_t * );
peer_t *search_friend ( peer_t *, struct in_addr, u_int16_t );
int     empty_neighbour_list ( void );
peer_t *is_successor ( peer_t * );
peer_t *is_neighbour ( unsigned long, u_int16_t );
peer_t *is_neighbour_or_finger ( peer_t * );
int     announce_friend
          ( peer_t *, struct in_addr, u_int16_t, unsigned int, char *, char *,
           pthread_t );

neblo_addr finger_distance_rule ( unsigned int magnitude );

peer_t *next_hop ( neblo_addr );

int     forward_msg
          ( rwd_t, peer_t *, neblo_port, neblo_addr, neblo_addr,
           char *, unsigned int, neblo_rhandle, int, pthread_t );

int     invoke_receiver_handler_UD
          ( neblo_port, neblo_addr, void *, unsigned int );
int     invoke_receiver_handler_REQ
          ( neblo_port, neblo_addr, void *, unsigned int * );
int     invoke_forwarder_handler_UD
          ( neblo_port, neblo_addr, void *, unsigned int );
int     invoke_forwarder_handler_REQ
          ( neblo_port, neblo_addr, void *, unsigned int );
int     invoke_forwarder_handler_RES
          ( neblo_port, neblo_addr, void *, unsigned int );
int     invoke_refresh_handler ( neblo_addr,  neblo_addr,  neblo_addr );
int     invoke_clean_handler ( neblo_addr, neblo_addr );

void    _close_ ( csocket_t ** );
void    _openout_ ( struct in_addr, u_int16_t, csocket_t **, RSA *, int * );
void    _openin_ ( u_int32_t, u_int16_t *, csocket_t **, int * );
void    _accept_
          ( csocket_t *, struct sockaddr *, socklen_t *, csocket_t **, int * );
void    _establishin_ ( csocket_t *, RSA *, RSA *, FILE *, char *, int * );
void    __read__ ( csocket_t *, void *, unsigned int, int *, long );
void    __write__ ( csocket_t *, void *, unsigned int, int * );

int     _put_challenge_ (csocket_t *, char *, RSA * );
int     _get_challenge_ (csocket_t *, char * );

void    cmd_handle_init ( cmd_handle_t * );
void    cmd_handle_set ( cmd_handle_t * );
#define cmd_handle_reset(h) do { (h)->val = 0; } while (0)
int     cmd_handle_age ( cmd_handle_t * );
#define cmd_handle_val(h) ((h)->val)
#define cmd_handle_lock(h) do { pthread_mutex_lock(&(h)->lock); } while (0)
#define cmd_handle_unlock(h) do { pthread_mutex_unlock(&(h)->lock); } while (0)
#define cmd_handle_is_done(h) ((h)->val == 0)


#endif
