/*********************************************************************/
/* findfinger.c                                                      */
/* function neblo_findfinger()                                       */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


/*
 *  findfinger <m> <scale_factor> <addr> <originator_ipaddr> <originator_port>
 *              <cmd_handle> <originator_keymod> <originator_keyexp>
 *
 *  remote peer <originator> wishes to meet its finger of magnitude <m>.
 *
 *  This req is managed in a recursive style (through a chain of recipients),
 *  so a <cmd_handle> is needed for the last recipient to answer back to
 *  the <originator>.
 *
 *  Let F(m) denote the finger of magnitude m.
 *
 *  Case m = 0 :
 *  distance between <originator> and its F(0) must be as large as possible
 *  but <= CUTOFF.  A "scale factor" is provided as an additional argument,
 *  to allow managing unpopulated rings, where the successor of a peer might
 *  be at > CUTOFF distance (and thus F(0) might be missing).
 *  For this distance to be evaluated precisely, the <originator> should
 *  disclose its own <addr> interval, what would be bad in terms of security.
 *  Rather, the <originator> yields an obfuscated <addr> by subtracting its own
 *  "smoke" (the "smoke" is subtracted from "my_id"; see finger_create(), in
 *  check_routes.c).  The recipient in turn subtracts its own "smoke",
 *  then evaluates the distance D between <addr> and the addr intvl of its
 *  successor (the successor addr interval is adjacent to the local interval so
 *  this can be done precisely).
 *  If D <= CUTOFF, then the recipient forwards the req to its successor.
 *  NOTE:  this does not reveal the addr interval of the recipient to its
 *  successor.
 *  The overall behaviour is that the request travels along a chain of
 *  successors until a peer P is found whose distance from originator is
 *  comprised between CUTOFF/2 and CUTOFF; the exact distance remains unknown.
 *  Peer P will become the finger F(0) of the <originator>, and will use
 *  the <cmd_handle> to directly contact the <originator>.
 *  The finger is inaccurate, but will work, since the real distance is smaller
 *  than the ideal value (namely, CUTOFF).
 *
 *  Case 1 <= m <= MAXFNGMAGN :
 *  In this case, the <addr> is always 1 or 0 and has a "depth" meaning as
 *  explained below.  The scale factor is always 0 and unused.
 *  For security reasons, the recipient will only manage a request like this
 *  from a remote peer P if it already got and answered a
 *  "findfinger" req of magnitude m-1 originated from P itself; that is,
 *  the recipient knows to be F(m-1) of P (this imposes that finger tables are
 *  built inductively; it is not a limitation, as the law for finger distances
 *  is recursive).
 *  The <originator> always issues a "findfinger" req with <addr>=1.
 *  When accepting a req with <addr>=1 (that is, directly from the
 *  <originator>), the recipient decreases the <addr> to 0, then forwards the
 *  req to its F(m-1).
 *  When accepting a req with <addr>=0, however, the recipient has to answer
 *  back to the <originator>, as the recipient can only be the finger of
 *  magnitude m of the <originator>; it moreover takes note that it has
 *  become finger F(m) of <originator>.
 *
 *  This is OK if the distance law is a geometric progression with base 2;
 *  for a Fibonacci law, req should be forwarded to F(m-1) which in turn
 *  should forward to its own F(m-2) which should answer back;
 *  this can be done by setting the first value of <addr> to 2 instead of 1
 *  (we clearly need a slightly different security policy on req acceptance,
 *  based on the current <addr> value).
 *
 *  A consequence of inaccurate F(0) and recursive building of
 *  subsequent F(M) is that the more distant, the more inaccurate a finger
 *  will be, although the inaccuracy is always a fixed fraction of the
 *  distance.  This is "the fog" of NEBLO.
 */

int
neblo_findfinger
( reqst_t *reqst )
{

  unsigned int          magn;
  unsigned int          scale_fact;
  neblo_addr            addr;
  unsigned long         raddr;
  struct in_addr        originator_addr;
  unsigned int          rport;
  u_int16_t             originator_port;
  u_int32_t             cmd_handle_val;
  char                 *originator_keymod;
  char                 *originator_keyexp;
  RSA                  *originator_pubkey = NULL;
  neblo_addr            dist;
  csocket_t            *cs = NULL;
  peer_t               *f;
  int                   rc;
  unsigned int          must_answer = 0;


  _do_(_malloc_(originator_keymod,KEYBUFLEN+SLACK),
       !originator_keymod, "neblo_findfinger(): originator_keymod malloc");
  _do_(_malloc_(originator_keyexp,BUFLEN+SLACK),
       !originator_keyexp, "neblo_findfinger(): originator_keyexp malloc");

  sscanf(reqst->buf,"%*s %u %u %016llx %lu %u %u %50000s %100s",
         &magn,&scale_fact,&addr,&raddr,&rport,&cmd_handle_val,
         originator_keymod,originator_keyexp);

  originator_addr.s_addr = raddr;
  originator_port = (u_int16_t)rport;

  /*  SANITY checks  */

  /*  Discard if <originator> is myself (entire ring traversed).  */

  if (originator_addr.s_addr == my_ipaddr.s_addr && originator_port == PORT) {

#ifdef DEBUG_R
    S(fprintf(stderr,"[%lu]\toriginator is myself. discard.\n",
              reqst->th_id));
#endif

    rc = -1;
    goto exit;
  }

  /*  SANITY checks OK. */

  if (magn == 0) {

    /*  SECURITY checks:  */

    /*
     *  Req may only come from our predecessor:  the peer that welcomed us
     *  in the ring.
     */
    if (!from_predecessor(reqst->csock)) {

#ifdef DEBUG_R
      S(fprintf(stderr,"[%lu]\tnot from predecessor. abort.\n",reqst->th_id));
#endif

      goto abort;
    }

    /*  reqsts for "excessive" distances are malicious!  */
    dist = _distance_(addr,ring->my_id);
    if (dist > (CUTOFF << scale_fact)) {

#ifdef DEBUG_R
      S(fprintf(stderr,
                "[%lu]\tdistance %016llx > scaled cutoff (%016llx). abort.\n",
                reqst->th_id,dist,CUTOFF << scale_fact));
#endif

      /*
       *  We can shut down the incoming connection.  This would indeed reveal
       *  information about our local addr interval, which is however already
       *  known by the predecessor, so don't worry.
       */
      goto abort;
    }

    /*  reqsts with "excessive" scale factor are malicious!  */
    if (scale_fact > 0 && dist <= (CUTOFF << (scale_fact-1))) {

#ifdef DEBUG_R
      S(fprintf(stderr,"[%lu]\tscale factor too large. abort.\n",
                reqst->th_id));
#endif

      goto abort;
    }

    /*  SECURITY checks OK. */

    /*
     *  Compute the _presumable_ distance between <addr> and successor, to
     *  decide a possible forwarding of this request to the successor peer.
     *  The <addr> must be obfuscated by local "smoke" before taking any
     *  forwarding decision.
     */
    addr -= ring->smoke;
    dist = _distance_(addr,ring->up_bound+1);

#ifdef DEBUG_R
    S(fprintf(stderr,
              "[%lu]\tdistance wrt successor is %016llx (cutoff at %016llx)\n",
              reqst->th_id,dist,CUTOFF));
#endif

    if (dist <= CUTOFF) {

      /* Successor presumably within CUTOFF, so forward to successor */

      f = successor();
      if (f == NULL)  /* should not happen */ {
        goto exit;
      }
      rc = tell_peer(f,reqst->buf,strlen(reqst->buf)+1,0,reqst->th_id);
      if (rc < 0)
        goto exit;

#ifdef DEBUG_R
      S(fprintf(stderr,"[%lu]\t<< %s\n",reqst->th_id,reqst->buf));
#endif

    } else {
      /*
       *  We are the last peer in the recipient chain to be within CUTOFF
       *  distance, so we are finger F(0) of <originator>.
       */
      must_answer = 1;
      /*  Apply the scale factor to get the actual magnitude.  */
      magn = scale_fact;
    }

  } else {

    /*  Magnitude is > 0  */

    /*  SECURITY/SANITY checks:  */

    /*
     *  Accept req with this magn only from friends already stamped "magn-1".
     *  The effect is twofold:  block requests that have travelled for more
     *  than the entire ring, and prevents any peer from knowing the entire
     *  finger table of local peer.
     */
    f = is_friend(reqst->csock);
    if (f == NULL) {

#ifdef DEBUG_R
      S(fprintf(stderr,"[%lu]\tnot a friend. abort.\n",reqst->th_id));
#endif

      goto abort;
    }
    if (f->magn != magn-1) {

#ifdef DEBUG_R
      S(fprintf(stderr,"[%lu]\toriginator has magn %u not %u. discard.\n",
                reqst->th_id,f->magn,magn-1));
#endif

      rc = -1;
      goto exit;
    }

    /*
     *  Accept req with magn > 0 only if <addr> is 1 or 0.
     *  To be changed if changing from a geometrical law to a Fibonacci law.
     */
    if (addr > 1)
      goto abort;

    /*  SECURITY/SANITY checks OK. */

    if (addr == 1) {

      /*  Forward to local finger F(magn-1), if it exists  */

      f = &ring->finger_table[magn-1];
      addr--;
      sprintf(reqst->buf,"findfinger %u %u %016llx %lu %u %u %s %s",
              magn,scale_fact,addr,raddr,rport,cmd_handle_val,
              originator_keymod,originator_keyexp);
      rc = tell_peer(f,reqst->buf,strlen(reqst->buf)+1,0,reqst->th_id);
      if (rc < 0) {
        /*  Failed or non-existent, and no alternatives.  Bail out.  */

#ifdef DEBUG_R
        S(fprintf(stderr,
                  "[%lu]\tneed finger of magn. %u to proceed. discard.\n",
                  reqst->th_id,magn-1));
#endif

        goto exit;
      }

#ifdef DEBUG_R
      S(fprintf(stderr,"[%lu]\t<< %s\n",reqst->th_id,reqst->buf));
#endif

    } else {
      /*
       *  <addr> = 0, that is, we are the final recipient.
       *  answer back to <originator>
       */
      must_answer = 1;
    }

  }

  if (must_answer) {

    /*
     *  Prepare to answer back to the original <originator>, as we are to
     *  become one of its fingers.  Send it a "newfinger" cmd.
     */

    _do_(_RSA_new_(originator_pubkey),!originator_pubkey,
         "neblo_findfinger(): RSA_new of originator_pubkey");
    _convert_pubkey_(originator_pubkey,originator_keymod,originator_keyexp,
                     goto exit);
    _openout_(originator_addr,originator_port,&cs,originator_pubkey,&rc);
    if (rc < 0)
      goto exit;

    /*  Prepare to get a friendship request from <originator> */
    insert_expd_reqst(originator_addr,originator_port,0 /* not predecessor */,
                      originator_pubkey,reqst->th_id);

    /*  Talk with <originator> at last!  */

    sprintf(reqst->buf,"newfinger %u %u %u %s %s",
            cmd_handle_val,magn,PORT,pubkeymod,pubkeyexp);
    _write_(cs,reqst->buf,strlen(reqst->buf)+1,rc,goto exit);

#ifdef DEBUG_R
    S(fprintf(stderr,"[%lu]\t<< %s\n",reqst->th_id,reqst->buf));
#endif

    /*
     *  This is needed to find <originator> in friend list when it will
     *  be inserted there.  See the search_friend() below.
     */
    f = ring->friend_list_tail;

    /*  ...challenge originator on his supposed pubkey... */
    rc = _put_challenge_(cs,reqst->buf,originator_pubkey);
    if (rc < 0) {

#ifdef DEBUG_R
      S(fprintf(stderr,"\tchallenge interrupted\n"));
#endif

      goto exit;
    }

#ifdef DEBUG_R
    S(fprintf(stderr,"\tchallenge OK\n"));
#endif

    /*  ...wait for originator to establish a friendship with us...  */
    _read_(cs,reqst->buf,BUFLEN,rc,TIMEOUT,goto exit);
    if (strncmp(reqst->buf,"ok",strlen("ok"))) {
      rc = -1;
      goto exit;
    }

#ifdef DEBUG
    S(fprintf(stderr,"[%lu]\t>> ok\n",reqst->th_id));
#endif

    /*
     *  Friendship has been established.  Search it in the tail of
     *  friend list, take note that it had catched us with a "findfinger" of
     *  magnitude <magn>, so we can only accept a "findfinger" of
     *  magnitude <magn>+1 from it.
     */
    f = search_friend(f->next,originator_addr,originator_port);
    if (f == NULL)  /* should not happen */
      goto exit;

    f->magn = magn;

  }

#ifdef DEBUG_R
  S(fprintf(stderr,"[%lu]\tDone.\n",reqst->th_id));
#endif

  rc = 0;

exit:

  _close_(&cs);
  _free_(originator_keymod); _free_(originator_keyexp);
  _RSA_free_(originator_pubkey);

  return rc;

abort:

  _close_(&reqst->csock);
  rc = -1;
  goto exit;


}
