/*********************************************************************/
/* join.c                                                            */
/* function neblo_join()                                             */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


/****************************************************************************

"join <requestor_id> <requestor_id_src> <requestor_ipaddr> <requestor_port>
<cmd_handle> <requestor_keymod> <requestor_keyexp>"

connect algorithm, phase 2.
Let T be the foreign node, ID(T) its id., S this peer:
If ID(T) falls into the interval of addresses stored at S, then S welcomes T
in the ring.
Otherwise, S applies the routing rule (neighs, and fingers) to ID(T), so as to
find another peer P who might be a better candidate for welcoming T;
then S forwards the join request to P.
If ID(T) is not "genuine" (it doesn't match its source through RIPEMD16), S
refuses to manage the request.

This req is generally managed in a recursive style (through a chain of
recipients), so a <cmd_handle> is needed for the last recipient to answer
back to the original <requestor>.

*****************************************************************************/


int
neblo_join
( reqst_t *reqst )
{

  neblo_addr            foreign_id;
  neblo_addr            foreign_id_src;
  struct in_addr        foreign_addr;
  unsigned long         faddr;
  u_int16_t             foreign_port;
  unsigned int          fport;
  u_int32_t             foreign_join_handle_val;
  char                 *foreign_keymod;
  char                 *foreign_keyexp;
  RSA                  *foreign_pubkey;
  unsigned char        *id_k;
  neblo_addr            id_chk;
  int                   rc = 0;
  csocket_t            *cs;
  unsigned int          i;
  peer_t                *f,*hop;
  neblo_addr             new_up_bound,foreign_up_bound;
  ringofptrs_t          *alt_ring;
  peerptr_t             *tail,*ff;
  char                 *neigh_keymod;
  char                 *neigh_keyexp;
  neblo_addr            old_up_bound;
  unsigned int          tmpunavail = 0;


  _do_(_malloc_(foreign_keymod,KEYBUFLEN+SLACK),
       !foreign_keymod, "neblo_join(): foreign_keymod malloc");
  _do_(_malloc_(foreign_keyexp,BUFLEN+SLACK),
       !foreign_keyexp, "neblo_join(): foreign_keyexp malloc");

  sscanf(reqst->buf,"%*s %016llx %016llx %lu %u %u %50000s %100s",
         &foreign_id,&foreign_id_src,&faddr,&fport,&foreign_join_handle_val,
         foreign_keymod,foreign_keyexp);

  foreign_addr.s_addr = faddr;

  /*  SECURITY  */

  foreign_port = (u_int16_t)fport;
  if ((unsigned int)foreign_port != fport) {

#ifdef DEBUG
    S(fprintf(stderr,"[%lu]\tneblo_join(): bad foreign port. Abort...\n",
              reqst->th_id));
#endif

    _free_(foreign_keymod); _free_(foreign_keyexp);
    return -1;
  }

  /* SECURITY II:  check foreign id against its source */

  _do_(_malloc_(id_k,RIPEMD160_DIGEST_LENGTH+SLACK), !id_k,
       "neblo_join(): id_k malloc");
  RIPEMD160((unsigned char *)&foreign_id_src,sizeof(foreign_id_src),id_k);
  xor_fold((int8_t *)&id_chk,sizeof(id_chk),
           (int8_t *)id_k,RIPEMD160_DIGEST_LENGTH);
  _free_(id_k);
  if (id_chk != foreign_id) {

#ifdef DEBUG_J
    S(fprintf(stderr,"[%lu]\tneblo_join(): bad foreign id. Abort...\n",
              reqst->th_id));
#endif

    _free_(foreign_keymod); _free_(foreign_keyexp);
    return -1;
  }

  /* SECURITY III:  ring must be already enabled, refuse otherwise */

  if (!ring->enabled) {

#ifdef DEBUG_J
    S(fprintf(stderr,"[%lu]\tneblo_join(): ring not enabled yet. Abort...\n",
              reqst->th_id));
#endif

    _free_(foreign_keymod); _free_(foreign_keyexp);
    return -1;
  }

  /* from now on, mutual exclusion w.r.t. other joins */
  pthread_mutex_lock(&ring->join_lock);

  for (;;) {

    /*
     *  Find a LIVING peer P whose id. has minimal distance from
     *  the foreign_id in the ring ordering.
     *  Local peer is included in the search.
     */

    if ((hop = next_hop(foreign_id)) == NULL) {

#ifdef DEBUG_J
      S(fprintf(stderr,"[%lu]\tclosest is myself\n",reqst->th_id));
#endif

      /*
       *  Local peer is the right one.  Must welcome foreign.
       *  The foreign peer becomes our new successor in the ring.
       * -) pass our current neighbour list to foreign peer;
       * -) adjust our neighbour list to include foreign peer as new successor
       *    (this may imply discarding the rightmost entry of neighbour list,
       *    if full);
       * -) divide my addr interval in two parts and assign second part
       *    to the foreign; second part starts from foreign id. so the two
       *    parts usually have different size;
       * -) get pubkey from the foreign peer, and challenge;
       * -) open a connection towards new successor, get a connection from it
       *    as well.
       */

      if (tmpunavail)
        forall_peers(f,f->tmpunavail = 0);

      /*  Open a csock and talk to foreign, to welcome it.  */

      _do_(_RSA_new_(foreign_pubkey),!foreign_pubkey,
           "neblo_join(): RSA_new of foreign_pubkey");
      _convert_pubkey_(foreign_pubkey,foreign_keymod,foreign_keyexp,rc = -1);
      if (rc < 0) {

#ifdef DEBUG_J
        S(fprintf(stderr,"[%lu]\tneblo_join(): bad foreign pubkey\n",
                  reqst->th_id));
#endif

        _RSA_free_(foreign_pubkey);
        rc = -1;
        break;
      }

      /*
       * Build a list of ptr to peers currently belonging to neigh list.
       * This will allow us to let the neigh list be concurrently changed
       * while this routine is still running.
       *
       * NOTE: this introduces aliasing on all peers in neigh list.
       *
       * NOTE: this temp copy ensures safety w.r.t. addition of new entries in
       * neigh list concurrently performed by check_routes().  However,
       * due to the aforementioned aliasing, we might still be unsafe w.r.t.
       * deletion of entries operated by check_routes() itself
       * (deletion is followed by free() which is harmful in case of aliasing).
       * That's why ("that's why"...) the "clean_lock" has been raised before
       * starting this service (see neblo.c)
       */
      _do_(_calloc_(alt_ring,sizeof(ringofptrs_t)),!alt_ring,
           "neblo_join(): alt_ring malloc");

      pthread_mutex_lock(&ring->lock);
      memcpy(&alt_ring->ring,ring,sizeof(ring_t));
      pthread_mutex_unlock(&ring->lock);

      alt_ring->peerlist = tail = NULL;

      forall_neighbours(f,
        _do_(_calloc_(ff,sizeof(peerptr_t)),!ff,"neblo_join(): ff calloc");
        ff->f = f;
        if (f == ring->neigh_list)
          alt_ring->peerlist = tail = ff;
        else {
          tail->next = ff;
          tail = ff;
        }
        ff->next = NULL;
      );

      /* now, welcome the foreign node */

      _openout_(foreign_addr,foreign_port,&cs,foreign_pubkey,&rc);
      if (rc < 0) {

#ifdef DEBUG_J
        S(fprintf(stderr,"[%lu]\tneblo_join(): refused conn from foreign\n",
                  reqst->th_id));
#endif

        goto abort;
      }

      sprintf(reqst->buf,"welcome %u %u %s %s",
              foreign_join_handle_val,PORT,pubkeymod,pubkeyexp);
      _write_(cs,reqst->buf,strlen(reqst->buf)+1,rc,goto abort);

#ifdef DEBUG_J
      S(fprintf(stderr,"[%lu]\t<< %s\n",reqst->th_id,reqst->buf));
#endif

      /*  Challenge foreign on his supposed pubkey. */
      rc = _put_challenge_(cs,reqst->buf,foreign_pubkey);
      if (rc < 0)
        goto abort;

#ifdef DEBUG_J
      S(fprintf(stderr,"[%lu]\tchallenge ok\n",reqst->th_id));
#endif

      /*
       * the addr interval assigned to this peer
       * splits in two parts: first part remains allocated
       * here, whereas second part becomes allocated to the foreign,
       * because it has greater id. in the 64-bit ring ordering
       */

      foreign_up_bound = ring->up_bound;
      new_up_bound = foreign_id - 1;

      sprintf(reqst->buf,"intvl %016llx",foreign_up_bound);
      _write_(cs,reqst->buf,strlen(reqst->buf)+1,rc,goto abort);

#ifdef DEBUG_J
      S(fprintf(stderr,"[%lu]\t<< %s\n",reqst->th_id,reqst->buf));
#endif

      /*  Send our current neighbour list.  */

      for (ff=alt_ring->peerlist; ff!=NULL; ff=ff->next) {

        f = ff->f;

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\tintroducing %u.%u.%u.%u:%u, status %u\n",
                  reqst->th_id,
                  ((unsigned char *)&f->addr.s_addr)[0],
                  ((unsigned char *)&f->addr.s_addr)[1],
                  ((unsigned char *)&f->addr.s_addr)[2],
                  ((unsigned char *)&f->addr.s_addr)[3],
                  f->port,f->status));
#endif

        if (!f->self) {
          /*
           * before introducing a neighbour "f" to the foreign,
           * announce to "f" that it may expect a friendship request
           * from the foreign
           */
          if (is_successor(f))
            rc = announce_friend(f,foreign_addr,foreign_port,
                                 1 /* will become its new predecessor */,
                                 foreign_keymod,foreign_keyexp,
                                 reqst->th_id);
          else
            rc = announce_friend(f,foreign_addr,foreign_port,
                                 0 /* just friend, not predecessor */,
                                 foreign_keymod,foreign_keyexp,
                                 reqst->th_id);
          if (rc < 0) {
            /*
             * something went wrong with neighbour "f".
             * "f" might have been flagged as "tmpunavail"
             */
            tmpunavail = 1;

#ifdef DEBUG_J
            S(fprintf(stderr,"[%lu]\t(neigh %u.%u.%u.%u:%u has troubles...)\n",
                      reqst->th_id,
                      ((unsigned char *)&f->addr.s_addr)[0],
                      ((unsigned char *)&f->addr.s_addr)[1],
                      ((unsigned char *)&f->addr.s_addr)[2],
                      ((unsigned char *)&f->addr.s_addr)[3],
                      f->port));
#endif

            /*
             *  If the faulting peer is our current successor we cannot
             *  welcome any other peer until a new successor is found.
             */
            if (ff == alt_ring->peerlist)
              break;
            else
              rc = 0;
          }
        } else {
          /*
           *  This covers the startup case, when this peer is the generator of
           *  the ring.
           *  When the peer first generated the ring, it had to include itself
           *  in its own neighbour list, to "keep the ring closed".
           *  As a consequence, each new successor get this peer in its
           *  neighbour list, so it will eventually open another conn. to us.
           */
          if (!empty_neighbour_list())
            insert_expd_reqst(foreign_addr,foreign_port,
                              0 /* new peer is successor, not predecessor */,
                              foreign_pubkey,reqst->th_id);
          else
            insert_expd_reqst(foreign_addr,foreign_port,
                              1 /*
                                 * empty ring: new peer is successor and
                                 * predecessor at the same time
                                 */,
                              foreign_pubkey,reqst->th_id);
        }

        _do_(_BN_bn2hex_(neigh_keymod,f->pubkey->n),!neigh_keymod,
             "neblo_join(): neigh_keymod alloc");
        _do_(_BN_bn2hex_(neigh_keyexp,f->pubkey->e),!neigh_keyexp,
             "neblo_join(): neigh_keyexp alloc");
        sprintf(reqst->buf,"%lu %u %s %s",
                (unsigned long)f->addr.s_addr,f->port,
                neigh_keymod,neigh_keyexp);
        _OPENSSL_free_(neigh_keymod); _OPENSSL_free_(neigh_keyexp);

        _write_(cs,reqst->buf,strlen(reqst->buf)+1,rc,break);

#ifdef DEBUG_J
        S(fprintf(stderr,"[%lu]\t<< \"neighbour %u.%u.%u.%u:%u\"\n",
                  reqst->th_id,
                  ((unsigned char *)&f->addr.s_addr)[0],
                  ((unsigned char *)&f->addr.s_addr)[1],
                  ((unsigned char *)&f->addr.s_addr)[2],
                  ((unsigned char *)&f->addr.s_addr)[3],
                  f->port));
#endif

        bzero(reqst->buf,4);
        _read_(cs,reqst->buf,BUFLEN,rc,TIMEOUT,break);
        if (strncmp(reqst->buf,"ok",strlen("ok"))) {

#ifdef DEBUG_J
          S(fprintf(stderr,"[%lu]\terror while welcoming new peer\n",
                  reqst->th_id));
#endif

          rc = -1;
          break;
        } else {

#ifdef DEBUG
          S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        }

      }

      if (rc < 0)
        /* something went wrong while talking to the foreign peer */
        goto abort;

      _write_(cs,"end",strlen("end")+1,rc,goto abort);

#ifdef DEBUG
      S(fprintf(stderr,"[%lu]\t<< end\n",reqst->th_id));
#endif

      /*
       *  Foreign node is now about to be inserted into the neighbour list,
       *  as the new successor.
       */

      _do_(_calloc_(f,sizeof(peer_t)),!f,"neblo_join(): f calloc");
      _pthmutexinit_(&f->lock);
      f->addr.s_addr = foreign_addr.s_addr;
      f->port = foreign_port;
      f->pubkey = foreign_pubkey;
      /*  All other fields are zero/NULL, by calloc  */

      /*
       *  Establish a permanent connection to the new successor
       *  for routing (become a "friend" of foreign).
       */

      /*  Wait for the new successor to be ready...  */
      bzero(reqst->buf,4);
      _read_(cs,reqst->buf,BUFLEN,rc,TIMEOUT,/* nothing */);
      if (strncmp(reqst->buf,"ok",strlen("ok"))) {
        /* something wrong while talking to the foreign peer.  abort... */

#ifdef DEBUG_J
        S(fprintf(stderr,"[%lu]\terror while waiting new peer to be ready\n",
                  reqst->th_id));
#endif

        goto abort2;
      } else {

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

      }

      f->status = TOCALL;

      for (i=5; i; i--) {
        rc = call_peer(f,1 /* from predecessor */,0,reqst->th_id);
        if (rc >= 0)
          /* success. status is set to OK */
          break;
        if (rc == -2) {
          /* new successor has died meanwhile. abort... */

#ifdef DEBUG_J
          S(fprintf(stderr,"[%lu]\tnew successor dead, abort...\n",
                    reqst->th_id));
#endif

          goto abort2;
        }
        /* new successor not responding. wait a moment, then try again */
        sleep(1);
      }
      if (!i) {
        /* new successor not responding for too long.  assume it is dead */

#ifdef DEBUG_J
          S(fprintf(stderr,"[%lu]\tnew successor not responding, abort...\n",
                    reqst->th_id));
#endif

          goto abort2;
      }

      /* in case still waiting an incoming conn. from previous successor... */
      remove_expd_reqst_from_successor(reqst->th_id);

      /* expect an incoming conn. from new successor */
      rc = insert_expd_reqst(f->addr,f->port,2 /* from successor */,f->pubkey,
                             reqst->th_id);

      /*
       *  Tell successor that we have connected to it and are ready to accept
       *  its own (backward) connection.
       */
      _write_(cs,"ok",strlen("ok")+1,rc,goto abort2);

#ifdef DEBUG_J
      S(fprintf(stderr,"[%lu]\t<< ok\n",reqst->th_id));
#endif

      /*
       *  The incoming conn from successor will be managed by a separate
       *  thread.  Successor tells us "ok" once it has connected.
       */
      bzero(reqst->buf,4);
      _read_(cs,reqst->buf,BUFLEN,rc,TIMEOUT,/* */);
      if (strncmp(reqst->buf,"ok",strlen("ok")))
        goto abort2;

#ifdef DEBUG
      S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

      /*  Two-way connection with successor has been established.  */

      /*
       *  Insert successor at head of our neighbour list, and
       *  update the local addr interval.
       */

      /*
       *  Inhibit msgs addressed to the local peer until done.
       *  This because the local addr intvl is about to change, and we don't
       *  want to take remote data to be then moved to the new peer.
       */
      pthread_mutex_lock(&ring->msg_lock);

      pthread_mutex_lock(&ring->lock);

      /*  Update addr interval  */
      old_up_bound = ring->up_bound;
      ring->up_bound = new_up_bound;

      /*  Insert at head  */
      f->next = ring->neigh_list;
      ring->neigh_list = f;

      pthread_mutex_unlock(&ring->lock);

      /*
       * Ask the upper layer (application) to migrate the data
       * mapped to the addr interval now assigned to the new peer
       * (primary as well as backup copies).
       */

      rc = republish_addrs(foreign_id,foreign_up_bound,foreign_id,MAX_CP);
      if (rc < 0) {

#ifdef DEBUG_J
        S(fprintf(stderr,"[%lu]\terror while republishing, abort join...\n",
                  reqst->th_id));
#endif

        goto abort4;
      }

      /* Everything was fine.  Final OK to the new successor. */

      _write_(cs,"ok",strlen("ok")+1,rc,/* */);
      if (rc < 0)
        /*  Something bad happened meanwhile (timeout?).  Abort...  */
        goto abort4;

#ifdef DEBUG_J
      S(fprintf(stderr,"[%lu]\t<< ok\n",reqst->th_id));
#endif

      /*  Almost done...  */

      /* resume accepting msgs */
      pthread_mutex_unlock(&ring->msg_lock);

      /*
       * now we can ask the upper layer (application) to remove the data
       * corresponding to the addr interval now migrated to the new peer
       */
      rc = clean_addrs(foreign_id,foreign_up_bound,MAX_CP);
      if (rc < 0) {

#ifdef DEBUG_J
        S(fprintf(stderr,"[%lu]\terror from clean handler, abort join...\n",
                  reqst->th_id));
#endif

        goto abort3;
      }

      /*  Foreign node entered ring successfully :-)  */

#ifdef DEBUG_J
      S(
      fprintf(stderr,"[%lu]\twelcomed a new peer:\n",
              reqst->th_id);
      fprintf(stderr,"[%lu]\tnew addr interval: %016llx -- %016llx\n",
              reqst->th_id,ring->my_id,ring->up_bound);
      fprintf(stderr,"[%lu]\tnew neighbour list:\n",reqst->th_id);
      forall_neighbours(f,
        if (f->status & (OK | TOCALL | PROBABLYDEAD))
          fprintf(stderr,"[%lu]\t\t%u.%u.%u.%u:%u (%u)\n",
                  reqst->th_id,
                  ((unsigned char *)&f->addr.s_addr)[0],
                  ((unsigned char *)&f->addr.s_addr)[1],
                  ((unsigned char *)&f->addr.s_addr)[2],
                  ((unsigned char *)&f->addr.s_addr)[3],
                  f->port,f->status);
      );
      );
#endif

      /* dispose the tmp copy of routing tables */
      for (ff=alt_ring->peerlist; ff!=NULL;) {
        tail = ff->next;
        _free_(ff);
        ff = tail;
      }
      _free_(alt_ring);

      rc = 0;
      break;

    }

    if (hop == (peer_t *)(-1)) {

#ifdef DEBUG_J
      S(fprintf(stderr,"[%lu]\tneblo_join(): couldn't route. Abort...\n",
                reqst->th_id));
#endif

      rc = -1;
      break;
    }

#ifdef DEBUG_J
    S(fprintf(stderr,"[%lu]\tnext hop is %u.%u.%u.%u:%u (status %u)\n",
              reqst->th_id,
              ((unsigned char *)&hop->addr.s_addr)[0],
              ((unsigned char *)&hop->addr.s_addr)[1],
              ((unsigned char *)&hop->addr.s_addr)[2],
              ((unsigned char *)&hop->addr.s_addr)[3],
              hop->port,hop->status));
#endif

    sprintf(reqst->buf,"join %016llx %016llx %lu %u %u %s %s",
            foreign_id,foreign_id_src,
            (unsigned long)foreign_addr.s_addr,(unsigned int)foreign_port,
            foreign_join_handle_val,
            foreign_keymod,foreign_keyexp);

    rc = tell_peer(hop,reqst->buf,strlen(reqst->buf)+1,1,reqst->th_id);

    if (rc < 0) {
      /*
       *  Peer at next hop had troubles, might have been flagged as
       *  "tmpunavail"
       */
      tmpunavail = 1;
      continue;
    }

#ifdef DEBUG_RW
    S(fprintf(stderr,"[%lu]\t(%u.%u.%u.%u:%u) << %s\n",
              reqst->th_id,
              ((unsigned char *)&hop->addr.s_addr)[0],
              ((unsigned char *)&hop->addr.s_addr)[1],
              ((unsigned char *)&hop->addr.s_addr)[2],
              ((unsigned char *)&hop->addr.s_addr)[3],
              hop->port,reqst->buf));
#endif


    break;


abort4:

    ring->up_bound = old_up_bound;
    ring->neigh_list = f->next;

abort3:

    /* resume accepting msgs */
    pthread_mutex_unlock(&ring->msg_lock);

abort2:

    clean_peer_desc(f);
    _free_(f);

abort:

    _RSA_free_(foreign_pubkey);

    /* dispose the tmp copy of routing tables */
    for (ff=alt_ring->peerlist; ff!=NULL;) {
      tail = ff->next;
      _free_(ff);
      ff = tail;
    }
    _free_(alt_ring);

    rc = -1;
    break;

  }

  _free_(foreign_keymod); _free_(foreign_keyexp);

  if (tmpunavail)
    forall_peers(f,f->tmpunavail = 0);

  /* resume accepting new join requests */
  pthread_mutex_unlock(&ring->join_lock);

  return rc;

}
