/*********************************************************************/
/* write.c                                                           */
/* function neblo_handle_UD_msgs()                                */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


/*
 * this peer is asked to take a write msg, and deliver to application
 * if this peer is the final destination, or forward on otherwise.
 */
int
neblo_handle_UD_msgs
( reqst_t *reqst, unsigned int retry )
{

  unsigned int          rport;
  neblo_port            port;
  unsigned int          len;
  neblo_addr            addr;
  neblo_addr            real_addr;
  int                   rc;
  peer_t               *hop;
  unsigned int          tmpunavail = 0;


  sscanf(reqst->buf,"%*s %04x %08x %016llx %016llx",
         &rport,&len,&addr,&real_addr);
  port = (neblo_port)rport;

#ifdef DEBUG_RW
  if (retry != FIRST_ATTMPT)
    S(fprintf(stderr,
              "[%lu]\tRETRY write: port %u, addr %016llx, real_addr %016llx, len %u, msg %s\n",
              reqst->th_id,port,addr,real_addr,len,reqst->buf+HEADERSIZE_W));
#endif

  /*
   * SECURITY: requests of this kind may only be accepted
   * if coming from a friend.
   * however, if this is a retry, the request might be from the
   * local peer; in this case it must be accepted
   */
  if (!is_friend(reqst->csock) && retry == FIRST_ATTMPT) {

#ifdef DEBUG_RW
    S(fprintf(stderr,"[%lu]\tREFUSED write: not a friend\n",reqst->th_id));
#endif
  
    return -1;
  }

  rc = 0;

  for (;;) {

    hop = next_hop(addr);

    if (hop == NULL) {
      /*
       * "addr" is mapped locally.
       * deliver msg to application, stripping header away
       */
      if (pthread_mutex_trylock(&ring->msg_lock)) {
        /*
         * A concurrent neblo_join() is in progress.
         * When finished, the local addr interval may have changed.
         * Therefore the route for this request must be recomputed.
         */
        /* wait for the flush to end... */
        pthread_mutex_lock(&ring->msg_lock);
        /* go recompute the route */
        pthread_mutex_unlock(&ring->msg_lock);
        continue;
      }
      rc = invoke_receiver_handler_UD(
             port,real_addr,reqst->buf+HEADERSIZE_W,len);
      pthread_mutex_unlock(&ring->msg_lock);

#ifdef DEBUG_RW
      S(fprintf(stderr,
                "[%lu]\t``write'' to port %u addr %016llx (real addr %016llx) returned %d, len %u, msg %s\n",
                reqst->th_id,port,addr,real_addr,rc,len,
                reqst->buf+HEADERSIZE_W));
#endif

      break;
    }

    if (hop == (peer_t *)(-1)) {
      /*
       * "addr" is not mapped here, and this peer doesn't know where to
       * route this msg copy at the moment.
       */

#ifdef DEBUG_RW
      if (retry != LAST_CHANCE)
        S(fprintf(stderr,
                  "[%lu]\twrite delayed: port %u, addr %016llx, real_addr %016llx, len %u, msg %s\n",
                  reqst->th_id,port,addr,real_addr,len,reqst->buf+HEADERSIZE_W
        ));
      else
        S(fprintf(stderr,
                  "[%lu]\twrite DISCARDED: port %u, addr %016llx, real_addr %016llx, len %u, msg %s\n",
                  reqst->th_id,port,addr,real_addr,len,reqst->buf+HEADERSIZE_W
        ));
#endif

      /*
       * insert the requested operation in a queue, to be tried again later,
       * but only if this is not a retry in itself (otherwise we could be
       * inserting the operation multiple times)
       */
      if (retry == FIRST_ATTMPT) {
        /*
         * we must do a copy. the original reqst buffer will be put back
         * in the service queue, the copy is to be put in the retry queue
         */
        reqst_t *r;

        _do_(_malloc_(r,sizeof(reqst_t)),!r,
             "neblo_handle_UD_msgs(): r malloc");
        memcpy(r,reqst,sizeof(reqst_t));
        _do_(_calloc_(r->buf,KEYBUFLEN+SLACK),!r->buf,
             "neblo_handle_UD_msgs(): r->buf calloc");
        memcpy(r->buf,reqst->buf,HEADERSIZE_W+len);
        insert_retry_queue(WRITE,r,reqst->th_id);
      }
      else
        if (retry == LAST_CHANCE) {
          /* no more chances.  Msg is dropped :-( */
        }

      rc = -1;
      break;
    }

    /* "addr" is mapped remotely; route msg through peer "hop" */

    rc = invoke_forwarder_handler_UD(port,real_addr,reqst->buf+HEADERSIZE_W,
                                     len);
    if (rc < 0)
      /*  application decided not to forward this msg  */
      break;

    rc = forward_msg(WRITE,hop,port,addr,real_addr,reqst->buf,len,0,1,
                     reqst->th_id);
    if (rc >= 0)
      /* success. quit service */
      break;

    /*
     * something wrong while talking to the selected peer, might have
     * been flagged as "tmpunavail".
     */

#ifdef DEBUG_RW
    S(fprintf(stderr,
              "[%lu]\tforward_msg() to addr %016llx (real addr %016llx) returned %d\n",
              reqst->th_id,addr,real_addr,rc));
#endif

    tmpunavail = 1;
    /* choose another peer */

  }

  /*
   * the forward_msg() routine above might have flagged some peers as
   * "tmpunavail"...
   */
  if (tmpunavail)
    forall_peers(f,f->tmpunavail = 0);

  return rc;

}
